import copy
import csv
import datetime
import logging
import os
import re
import tempfile
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import Dict, Iterable, List, Optional, Set, Tuple

import boto3
import requests
from botocore.exceptions import ClientError, EndpointConnectionError

from gpuhunt._internal.models import QueryFilter, RawCatalogItem
from gpuhunt.providers import AbstractProvider

logger = logging.getLogger(__name__)
ec2_pricing_url = (
    "https://pricing.us-east-1.amazonaws.com/offers/v1.0/aws/AmazonEC2/current/index.csv"
)
disclaimer_rows_skip = 5
# https://aws.amazon.com/ec2/previous-generation/
previous_generation_families = [
    "t1.",
    "m1.",
    "m3.",
    "c1.",
    "c3.",
    "i2.",
    "m2.",
    "cr1.",
    "r3.",
    "hs1.",
    "g2.",
    "a1.",
]
pricing_filters = {
    "TermType": ["OnDemand"],
    "Tenancy": ["Shared"],
    "Operating System": ["Linux"],
    "CapacityStatus": ["Used"],
    "Unit": ["Hrs"],
    "Currency": ["USD"],
    "Pre Installed S/W": ["", "NA"],
}
describe_instances_limit = 100


class AWSProvider(AbstractProvider):
    """
    AWSProvider parses Bulk API index file for AmazonEC2 in all regions and fills missing GPU details

    Required IAM permissions:
    * `ec2:DescribeInstanceTypes`
    """

    NAME = "aws"

    def __init__(self, cache_path: Optional[str] = None):
        if cache_path:
            self.cache_path = cache_path
        else:
            self.temp_dir = tempfile.TemporaryDirectory()
            self.cache_path = self.temp_dir.name + "/index.csv"
        # todo aws creds
        self.preview_gpus = {
            "p4de.24xlarge": ("A100", 80.0),
        }

    def get(self, query_filter: Optional[QueryFilter] = None) -> List[RawCatalogItem]:
        if not os.path.exists(self.cache_path):
            logger.info("Downloading EC2 prices to %s", self.cache_path)
            with requests.get(ec2_pricing_url, stream=True) as r:
                r.raise_for_status()
                with open(self.cache_path, "wb") as f:
                    for chunk in r.iter_content(chunk_size=8192):
                        f.write(chunk)

        offers = []
        with open(self.cache_path, "r", newline="") as f:
            for _ in range(disclaimer_rows_skip):
                f.readline()
            reader: Iterable[dict[str, str]] = csv.DictReader(f)
            for row in reader:
                if self.skip(row):
                    continue
                offer = RawCatalogItem(
                    instance_name=row["Instance Type"],
                    location=row["Region Code"],
                    price=float(row["PricePerUnit"]),
                    cpu=int(row["vCPU"]),
                    memory=parse_memory(row["Memory"]),
                    gpu_count=parse_optional_count(row["GPU"]),
                    spot=False,
                    gpu_name=None,
                    gpu_memory=None,
                )
                offers.append(offer)
        self.fill_gpu_details(offers)
        return self.add_spots(offers)

    def skip(self, row: Dict[str, str]) -> bool:
        if any(row["Instance Type"].startswith(family) for family in previous_generation_families):
            return True
        for key, values in pricing_filters.items():
            if row[key] not in values:
                return True
        return False

    def fill_gpu_details(self, offers: List[RawCatalogItem]):
        regions = defaultdict(list)
        for offer in offers:
            if offer.gpu_count > 0 and offer.instance_name not in self.preview_gpus:
                regions[offer.location].append(offer.instance_name)

        gpus = copy.deepcopy(self.preview_gpus)
        while regions:
            region = max(regions, key=lambda r: len(regions[r]))
            instance_types = regions.pop(region)

            client = boto3.client("ec2", region_name=region)
            paginator = client.get_paginator("describe_instance_types")
            for offset in range(0, len(instance_types), describe_instances_limit):
                logger.info("Fetching GPU details for %s (offset=%s)", region, offset)
                pages = paginator.paginate(
                    InstanceTypes=instance_types[offset : offset + describe_instances_limit]
                )
                for page in pages:
                    for i in page["InstanceTypes"]:
                        gpu = i["GpuInfo"]["Gpus"][0]
                        gpus[i["InstanceType"]] = (
                            gpu["Name"],
                            gpu["MemoryInfo"]["SizeInMiB"] / 1024,
                        )

            regions = {
                region: left
                for region, names in regions.items()
                if (left := [i for i in names if i not in instance_types])
            }

        for offer in offers:
            if offer.gpu_count > 0:
                offer.gpu_name, offer.gpu_memory = gpus[offer.instance_name]

    def _add_spots_worker(
        self, region: str, instance_types: Set[str]
    ) -> Dict[Tuple[str, str], float]:
        spot_prices = dict()
        logger.info("Fetching spot prices for %s", region)
        try:
            client = boto3.client("ec2", region_name=region)  # todo creds
            pages = client.get_paginator("describe_spot_price_history").paginate(
                Filters=[
                    {
                        "Name": "product-description",
                        "Values": ["Linux/UNIX"],
                    }
                ],
                InstanceTypes=list(instance_types),
                StartTime=datetime.datetime.utcnow(),
            )

            instance_prices = defaultdict(list)
            for page in pages:
                for item in page["SpotPriceHistory"]:
                    instance_prices[item["InstanceType"]].append(float(item["SpotPrice"]))
            for (
                instance_type,
                zone_prices,
            ) in instance_prices.items():  # reduce zone prices to a single value
                spot_prices[(instance_type, region)] = min(zone_prices)
        except (ClientError, EndpointConnectionError):
            return {}
        return spot_prices

    def add_spots(self, offers: List[RawCatalogItem]) -> List[RawCatalogItem]:
        region_instances = defaultdict(set)
        for offer in offers:
            region_instances[offer.location].add(offer.instance_name)

        spot_prices = dict()
        with ThreadPoolExecutor(max_workers=8) as executor:
            future_to_region = {}
            for region, instance_types in region_instances.items():
                future = executor.submit(self._add_spots_worker, region, instance_types)
                future_to_region[future] = region
            for future in as_completed(future_to_region):
                spot_prices.update(future.result())

        spot_offers = []
        for offer in offers:
            if (price := spot_prices.get((offer.instance_name, offer.location))) is None:
                continue
            spot_offer = copy.deepcopy(offer)
            spot_offer.spot = True
            spot_offer.price = price
            spot_offers.append(spot_offer)
        return offers + spot_offers

    @classmethod
    def filter(cls, offers: List[RawCatalogItem]) -> List[RawCatalogItem]:
        return [
            i
            for i in offers
            if any(
                i.instance_name.startswith(family)
                for family in [
                    "t2.small",
                    "c5.",
                    "m5.",
                    "p3.",
                    "g5.",
                    "g4dn.",
                    "p4d.",
                    "p4de.",
                ]
            )
        ]


def parse_memory(s: str) -> float:
    r = re.match(r"^([0-9.]+) GiB$", s)
    return float(r.group(1))


def parse_optional_count(s: str) -> int:
    if not s:
        return 0
    return int(s)



from dataclasses import asdict, dataclass
from typing import Dict, List, Optional, Tuple, Union

from gpuhunt._internal.utils import empty_as_none


@dataclass
class RawCatalogItem:
    instance_name: Optional[str]
    location: Optional[str]
    price: Optional[float]
    cpu: Optional[int]
    memory: Optional[float]
    gpu_count: Optional[int]
    gpu_name: Optional[str]
    gpu_memory: Optional[float]
    spot: Optional[bool]

    @staticmethod
    def from_dict(v: dict) -> "RawCatalogItem":
        return RawCatalogItem(
            instance_name=empty_as_none(v.get("instance_name")),
            location=empty_as_none(v.get("location")),
            price=empty_as_none(v.get("price"), loader=float),
            cpu=empty_as_none(v.get("cpu"), loader=int),
            memory=empty_as_none(v.get("memory"), loader=float),
            gpu_count=empty_as_none(v.get("gpu_count"), loader=int),
            gpu_name=empty_as_none(v.get("gpu_name")),
            gpu_memory=empty_as_none(v.get("gpu_memory"), loader=float),
            spot=empty_as_none(v.get("spot"), loader=lambda x: x.lower() == "true"),
        )

    def dict(self) -> Dict[str, Union[str, int, float, bool, None]]:
        return asdict(self)


@dataclass
class CatalogItem(RawCatalogItem):
    """
    Attributes:
        instance_name: name of the instance
        location: region or zone
        price: $ per hour
        cpu: number of CPUs
        memory: amount of RAM in GB
        gpu_count: number of GPUs
        gpu_name: name of the GPU
        gpu_memory: amount of GPU VRAM in GB for each GPU
        spot: whether the instance is a spot instance
        provider: name of the provider
    """

    instance_name: str
    location: str
    price: float
    cpu: int
    memory: float
    gpu_count: int
    gpu_name: Optional[str]
    gpu_memory: Optional[float]
    spot: bool
    provider: str

    @staticmethod
    def from_dict(v: dict, *, provider: Optional[str] = None) -> "CatalogItem":
        return CatalogItem(provider=provider, **asdict(RawCatalogItem.from_dict(v)))


@dataclass
class QueryFilter:
    """
    Attributes:
        provider: name of the provider to filter by. If not specified, all providers will be used
        min_cpu: minimum number of CPUs
        max_cpu: maximum number of CPUs
        min_memory: minimum amount of RAM in GB
        max_memory: maximum amount of RAM in GB
        min_gpu_count: minimum number of GPUs
        max_gpu_count: maximum number of GPUs
        gpu_name: case-sensitive name of the GPU to filter by. If not specified, all GPUs will be used
        min_gpu_memory: minimum amount of GPU VRAM in GB for each GPU
        max_gpu_memory: maximum amount of GPU VRAM in GB for each GPU
        min_total_gpu_memory: minimum amount of GPU VRAM in GB for all GPUs combined
        max_total_gpu_memory: maximum amount of GPU VRAM in GB for all GPUs combined
        min_disk_size: *currently not in use*
        max_disk_size: *currently not in use*
        min_price: minimum price per hour in USD
        max_price: maximum price per hour in USD
        min_compute_capability: minimum compute capability of the GPU
        max_compute_capability: maximum compute capability of the GPU
        spot: if `False`, only ondemand offers will be returned. If `True`, only spot offers will be returned
    """

    provider: Optional[List[str]] = None
    min_cpu: Optional[int] = None
    max_cpu: Optional[int] = None
    min_memory: Optional[float] = None
    max_memory: Optional[float] = None
    min_gpu_count: Optional[int] = None
    max_gpu_count: Optional[int] = None
    gpu_name: Optional[List[str]] = None
    min_gpu_memory: Optional[float] = None
    max_gpu_memory: Optional[float] = None
    min_total_gpu_memory: Optional[float] = None
    max_total_gpu_memory: Optional[float] = None
    min_disk_size: Optional[int] = None
    max_disk_size: Optional[int] = None
    min_price: Optional[float] = None
    max_price: Optional[float] = None
    min_compute_capability: Optional[Tuple[int, int]] = None
    max_compute_capability: Optional[Tuple[int, int]] = None
    spot: Optional[bool] = None


@dataclass
class GPUInfo:
    name: str
    memory: int
    compute_capability: Tuple[int, int]



import argparse
import logging
import os
import sys

import gpuhunt._internal.storage as storage


def main():
    parser = argparse.ArgumentParser(prog="python3 -m gpuhunt")
    parser.add_argument(
        "provider",
        choices=["aws", "azure", "gcp", "lambdalabs", "tensordock", "vastai"],
    )
    parser.add_argument("--output", required=True)
    parser.add_argument("--no-filter", action="store_true")
    args = parser.parse_args()
    logging.basicConfig(
        level=logging.INFO,
        stream=sys.stdout,
        format="%(asctime)s %(levelname)s %(message)s",
    )

    if args.provider == "aws":
        from gpuhunt.providers.aws import AWSProvider

        provider = AWSProvider(os.getenv("AWS_CACHE_PATH"))
    elif args.provider == "azure":
        from gpuhunt.providers.azure import AzureProvider

        provider = AzureProvider(os.getenv("AZURE_SUBSCRIPTION_ID"))
    elif args.provider == "gcp":
        from gpuhunt.providers.gcp import GCPProvider

        provider = GCPProvider(os.getenv("GCP_PROJECT_ID"))
    elif args.provider == "lambdalabs":
        from gpuhunt.providers.lambdalabs import LambdaLabsProvider

        provider = LambdaLabsProvider(os.getenv("LAMBDALABS_TOKEN"))
    elif args.provider == "tensordock":
        from gpuhunt.providers.tensordock import TensorDockProvider

        provider = TensorDockProvider()
    elif args.provider == "vastai":
        from gpuhunt.providers.vastai import VastAIProvider

        provider = VastAIProvider()
    else:
        exit(f"Unknown provider {args.provider}")

    logging.info("Fetching offers for %s", args.provider)
    offers = provider.get()
    if not args.no_filter:
        offers = provider.filter(offers)
    storage.dump(offers, args.output)


if __name__ == "__main__":
    main()



from gpuhunt._internal.catalog import Catalog
from gpuhunt._internal.constraints import fill_missing, matches
from gpuhunt._internal.default import default_catalog, query
from gpuhunt._internal.models import CatalogItem, QueryFilter, RawCatalogItem



from pathlib import Path

import pytest


@pytest.fixture
def data(catalog_dir: Path) -> str:
    return (catalog_dir / "aws.csv").read_text()


class TestAWSCatalog:
    def test_m5_large_regions(self, data: str):
        instance = "m5.large"
        regions = [
            "af-south-1",
            "ap-east-1",
            "ap-northeast-1",
            "ap-northeast-2",
            "ap-northeast-3",
            "ap-south-1",
            "ap-south-2",
            "ap-southeast-1",
            "ap-southeast-2",
            "ap-southeast-3",
            "ap-southeast-4",
            "ca-central-1",
            "eu-central-1",
            "eu-central-2",
            "eu-north-1",
            "eu-south-1",
            "eu-south-2",
            "eu-west-1",
            "eu-west-2",
            "eu-west-3",
            "il-central-1",
            "me-central-1",
            "me-south-1",
            "sa-east-1",
            "us-east-1",
            "us-east-2",
            "us-gov-east-1",
            "us-gov-west-1",
            "us-west-1",
            "us-west-2",
            "us-west-2-lax-1",
        ]
        assert all(f"\n{instance},{i}," in data for i in regions)

    def test_spots_presented(self, data: str):
        assert ",True\n" in data

    def test_gpu_presented(self, data: str):
        gpus = [
            "A100",
            "A10G",
            "T4",
            "V100",
        ]
        assert all(f",{i}," in data for i in gpus)



from typing import Callable, Optional, Tuple, Union


def empty_as_none(value: Optional[str], loader: Optional[Callable] = None):
    if value is None or value == "":
        return None
    if loader is not None:
        return loader(value)
    return value


def parse_compute_capability(
    value: Optional[Union[str, Tuple[int, int]]]
) -> Optional[Tuple[int, int]]:
    if isinstance(value, str):
        major, minor = value.split(".")
        return int(major), int(minor)
    return value



from abc import ABC, abstractmethod
from typing import List, Optional

from gpuhunt._internal.models import QueryFilter, RawCatalogItem


class AbstractProvider(ABC):
    NAME: str = "abstract"

    @abstractmethod
    def get(self, query_filter: Optional[QueryFilter] = None) -> List[RawCatalogItem]:
        pass

    @classmethod
    def filter(cls, offers: List[RawCatalogItem]) -> List[RawCatalogItem]:
        return offers



import copy
from typing import List, Optional

import requests

from gpuhunt._internal.models import QueryFilter, RawCatalogItem
from gpuhunt.providers import AbstractProvider

bundles_url = "https://console.vast.ai/api/v0/bundles"


class VastAIProvider(AbstractProvider):
    NAME = "vastai"

    def get(self, query_filter: Optional[QueryFilter] = None) -> List[RawCatalogItem]:
        data = requests.get(bundles_url).json()
        instance_offers = []
        for offer in data["offers"]:
            gpu_name = get_gpu_name(offer["gpu_name"])
            ondemand_offer = RawCatalogItem(
                instance_name=f"{offer['host_id']}",
                location=get_location(offer["geolocation"]),
                price=round(offer["dph_total"], 5),
                cpu=offer["cpu_cores"],
                memory=round(offer["cpu_ram"] / 1024),
                gpu_count=offer["num_gpus"],
                gpu_name=gpu_name,
                gpu_memory=round(offer["gpu_ram"] / 1024),
                spot=False,
            )
            instance_offers.append(ondemand_offer)

            spot_offer = copy.deepcopy(ondemand_offer)
            spot_offer.price = round(offer["min_bid"], 5)
            spot_offer.spot = True
            instance_offers.append(spot_offer)
        return instance_offers


def get_gpu_name(gpu_name: str) -> str:
    gpu_name = gpu_name.replace("RTX A", "A").replace("Tesla ", "").replace("Q ", "")
    if gpu_name.startswith("A100 "):
        return "A100"
    return gpu_name.replace(" ", "")


def get_location(location: Optional[str]) -> str:
    if location is None:
        return ""
    try:
        city, country = location.replace(", ", ",").split(",")
        location = f"{country}-{city}"
    except ValueError:
        pass
    return location.lower().replace(" ", "")



import csv
import dataclasses
from typing import Iterable, List, Type, TypeVar

from gpuhunt._internal.models import RawCatalogItem

T = TypeVar("T", bound=RawCatalogItem)


def dump(items: List[T], path: str, *, cls: Type[T] = RawCatalogItem):
    with open(path, "w", newline="") as f:
        writer = csv.DictWriter(f, fieldnames=[field.name for field in dataclasses.fields(cls)])
        writer.writeheader()
        for item in items:
            writer.writerow(item.dict())


def load(path: str, *, cls: Type[T] = RawCatalogItem) -> List[T]:
    items = []
    with open(path, "r", newline="") as f:
        reader: Iterable[dict[str, str]] = csv.DictReader(f)
        for row in reader:
            item = cls.from_dict(row)
            items.append(item)
    return items


def sort_key(offer: RawCatalogItem):
    return offer.gpu_count, offer.instance_name, offer.price, offer.location



import pytest

from gpuhunt import CatalogItem, QueryFilter
from gpuhunt._internal.constraints import fill_missing, matches


@pytest.fixture
def item() -> CatalogItem:
    return CatalogItem(
        instance_name="large",
        location="us-east-1",
        price=1.2,
        cpu=16,
        memory=64.0,
        gpu_count=1,
        gpu_name="A100",
        gpu_memory=40.0,
        spot=False,
        provider="aws",
    )


class TestMatches:
    def test_empty(self, item: CatalogItem):
        assert matches(item, QueryFilter())

    def test_cpu(self, item: CatalogItem):
        assert matches(item, QueryFilter(min_cpu=16))
        assert matches(item, QueryFilter(max_cpu=16))
        assert not matches(item, QueryFilter(min_cpu=32))
        assert not matches(item, QueryFilter(max_cpu=8))

    def test_memory(self, item: CatalogItem):
        assert matches(item, QueryFilter(min_memory=64.0))
        assert matches(item, QueryFilter(max_memory=64.0))
        assert not matches(item, QueryFilter(min_memory=128.0))
        assert not matches(item, QueryFilter(max_memory=32.0))

    def test_gpu_count(self, item: CatalogItem):
        assert matches(item, QueryFilter(min_gpu_count=1))
        assert matches(item, QueryFilter(max_gpu_count=1))
        assert not matches(item, QueryFilter(min_gpu_count=2))
        assert not matches(item, QueryFilter(max_gpu_count=0))

    def test_gpu_memory(self, item: CatalogItem):
        assert matches(item, QueryFilter(min_gpu_memory=40.0))
        assert matches(item, QueryFilter(max_gpu_memory=40.0))
        assert not matches(item, QueryFilter(min_gpu_memory=80.0))
        assert not matches(item, QueryFilter(max_gpu_memory=20.0))

    def test_gpu_name(self, item: CatalogItem):
        assert matches(item, QueryFilter(gpu_name=["A100"]))
        assert not matches(item, QueryFilter(gpu_name=["A10"]))

    def test_total_gpu_memory(self, item: CatalogItem):
        assert matches(item, QueryFilter(min_total_gpu_memory=40.0))
        assert matches(item, QueryFilter(max_total_gpu_memory=40.0))
        assert not matches(item, QueryFilter(min_total_gpu_memory=80.0))
        assert not matches(item, QueryFilter(max_total_gpu_memory=20.0))

    def test_price(self, item: CatalogItem):
        assert matches(item, QueryFilter(min_price=1.2))
        assert matches(item, QueryFilter(max_price=1.2))
        assert not matches(item, QueryFilter(min_price=1.3))
        assert not matches(item, QueryFilter(max_price=1.1))

    def test_spot(self, item: CatalogItem):
        assert matches(item, QueryFilter(spot=False))
        assert not matches(item, QueryFilter(spot=True))

    def test_compute_capability(self, item: CatalogItem):
        assert matches(item, QueryFilter(min_compute_capability=(8, 0)))
        assert matches(item, QueryFilter(max_compute_capability=(8, 0)))
        assert not matches(item, QueryFilter(min_compute_capability=(8, 1)))
        assert not matches(item, QueryFilter(max_compute_capability=(7, 9)))


class TestFillMissing:
    def test_empty(self):
        assert fill_missing(QueryFilter(), memory_per_core=4) == QueryFilter()

    def test_from_cpu(self):
        assert fill_missing(QueryFilter(min_cpu=2), memory_per_core=4) == QueryFilter(
            min_cpu=2,
            min_memory=8,
        )

    def test_from_memory(self):
        assert fill_missing(QueryFilter(min_memory=6), memory_per_core=4) == QueryFilter(
            min_memory=6,
            min_cpu=2,
        )

    def test_from_total_gpu_memory(self):
        assert fill_missing(
            QueryFilter(min_total_gpu_memory=24), memory_per_core=4
        ) == QueryFilter(
            min_total_gpu_memory=24,
            min_memory=48,
            min_disk_size=54,
            min_cpu=12,
        )

    def test_from_gpu_memory(self):
        assert fill_missing(QueryFilter(min_gpu_memory=16), memory_per_core=4) == QueryFilter(
            min_gpu_memory=16,
            min_memory=32,
            min_disk_size=46,
            min_cpu=8,
        )

    def test_from_gpu_count(self):
        assert fill_missing(QueryFilter(min_gpu_count=2), memory_per_core=4) == QueryFilter(
            min_gpu_count=2,
            min_memory=32,  # minimal GPU has 8 GB of memory
            min_disk_size=46,
            min_cpu=8,
        )

    def test_from_gpu_name(self):
        assert fill_missing(QueryFilter(gpu_name=["A100"]), memory_per_core=4) == QueryFilter(
            gpu_name=["A100"],
            min_memory=80,
            min_disk_size=70,
            min_cpu=20,
        )

    def test_from_compute_capability(self):
        assert fill_missing(
            QueryFilter(min_compute_capability=(9, 0)), memory_per_core=4
        ) == QueryFilter(
            min_compute_capability=(9, 0),
            min_memory=160,
            min_disk_size=110,
            min_cpu=40,
        )



import copy
import logging
import re
from typing import List, Optional, Tuple

import requests

from gpuhunt._internal.models import QueryFilter, RawCatalogItem
from gpuhunt.providers import AbstractProvider

logger = logging.getLogger(__name__)
instance_types_url = "https://cloud.lambdalabs.com/api/v1/instance-types"
all_regions = [
    "us-south-1",
    "us-west-2",
    "us-west-1",
    "us-midwest-1",
    "us-west-3",
    "us-east-1",
    "europe-central-1",
    "asia-south-1",
    "me-west-1",
    "asia-northeast-1",
    "asia-northeast-2",
]


class LambdaLabsProvider(AbstractProvider):
    NAME = "lambdalabs"

    def __init__(self, token: str):
        self.token = token

    def get(self, query_filter: Optional[QueryFilter] = None) -> List[RawCatalogItem]:
        offers = []
        data = requests.get(
            instance_types_url, headers={"Authorization": f"Bearer {self.token}"}
        ).json()["data"]
        for instance in data.values():
            instance = instance["instance_type"]
            logger.info(instance["name"])
            gpu_count, gpu_name, gpu_memory = parse_description(instance["description"])
            offer = RawCatalogItem(
                instance_name=instance["name"],
                price=instance["price_cents_per_hour"] / 100,
                cpu=instance["specs"]["vcpus"],
                memory=float(instance["specs"]["memory_gib"]),
                gpu_count=gpu_count,
                gpu_name=gpu_name,
                gpu_memory=gpu_memory,
                spot=False,
                location=None,
            )
            offers.append(offer)
        return self.add_regions(offers)

    def add_regions(self, offers: List[RawCatalogItem]) -> List[RawCatalogItem]:
        # TODO: we don't know which regions are actually available for each instance type
        region_offers = []
        for region in all_regions:
            for offer in offers:
                offer = copy.deepcopy(offer)
                offer.location = region
                region_offers.append(offer)
        return region_offers


def parse_description(v: str) -> Tuple[int, str, float]:
    """Returns gpus count, gpu name, and GPU memory"""
    r = re.match(r"^(\d)x (?:Tesla )?(.+) \((\d+) GB", v)
    count, gpu_name, gpu_memory = r.groups()
    return int(count), gpu_name.replace(" ", ""), float(gpu_memory)



from pathlib import Path

import pytest


@pytest.fixture
def data(catalog_dir: Path) -> str:
    return (catalog_dir / "azure.csv").read_text()


class TestAzureCatalog:
    def test_standard_d2s_v3_locations(self, data: str):
        locations = [
            "attatlanta1",
            "attdallas1",
            "attdetroit1",
            "attnewyork1",
            "australiacentral",
            "australiacentral2",
            "australiaeast",
            "australiasoutheast",
            "australiasoutheast",
            "brazilsouth",
            "brazilsoutheast",
            "canadacentral",
            "canadaeast",
            "centralindia",
            "centralus",
            "eastasia",
            "eastus",
            "eastus2",
            "francecentral",
            "francesouth",
            "germanynorth",
            "germanywestcentral",
            "israelcentral",
            "italynorth",
            "japaneast",
            "japanwest",
            "jioindiacentral",
            "jioindiawest",
            "koreacentral",
            "koreasouth",
            "northcentralus",
            "northeurope",
            "norwayeast",
            "norwaywest",
            "polandcentral",
            "qatarcentral",
            "sgxsingapore1",
            "southafricanorth",
            "southafricawest",
            "southcentralus",
            "southeastasia",
            "southindia",
            "swedencentral",
            "swedensouth",
            "switzerlandnorth",
            "switzerlandwest",
            "uaecentral",
            "uaenorth",
            "uksouth",
            "ukwest",
            "usgovarizona",
            "usgovtexas",
            "usgovvirginia",
            "westcentralus",
            "westeurope",
            "westindia",
            "westus",
            "westus2",
            "westus3",
        ]
        assert all(f"\nStandard_D2s_v3,{i}," in data for i in locations)

    def test_spots_presented(self, data: str):
        assert ",True\n" in data

    def test_ondemand_presented(self, data: str):
        assert ",False\n" in data

    def test_gpu_presented(self, data: str):
        gpus = [
            "A100",
            "A10",
            "T4",
            "V100",
        ]
        assert all(f",{i}," in data for i in gpus)

    def test_both_a100_presented(self, data: str):
        assert ",A100,40.0," in data
        assert ",A100,80.0," in data



from pathlib import Path

import pytest


@pytest.fixture
def data(catalog_dir: Path) -> str:
    return (catalog_dir / "gcp.csv").read_text()


class TestGCPCatalog:
    def test_e2_highcpu_2_zones(self, data: str):
        zones = [
            "asia-east1-a",
            "asia-east1-b",
            "asia-east1-c",
            "asia-east2-a",
            "asia-east2-b",
            "asia-east2-c",
            "asia-northeast1-a",
            "asia-northeast1-b",
            "asia-northeast1-c",
            "asia-northeast2-a",
            "asia-northeast2-b",
            "asia-northeast2-c",
            "asia-northeast3-a",
            "asia-northeast3-b",
            "asia-northeast3-c",
            "asia-south1-a",
            "asia-south1-b",
            "asia-south1-c",
            "asia-south2-a",
            "asia-south2-b",
            "asia-south2-c",
            "asia-southeast1-a",
            "asia-southeast1-b",
            "asia-southeast1-c",
            "asia-southeast2-a",
            "asia-southeast2-b",
            "asia-southeast2-c",
            "australia-southeast1-a",
            "australia-southeast1-b",
            "australia-southeast1-c",
            "australia-southeast2-a",
            "australia-southeast2-b",
            "australia-southeast2-c",
            "europe-central2-a",
            "europe-central2-b",
            "europe-central2-c",
            "europe-north1-a",
            "europe-north1-b",
            "europe-north1-c",
            "europe-southwest1-a",
            "europe-southwest1-b",
            "europe-southwest1-c",
            "europe-west1-b",
            "europe-west1-c",
            "europe-west1-d",
            "europe-west10-a",
            "europe-west10-b",
            "europe-west10-c",
            "europe-west12-a",
            "europe-west12-b",
            "europe-west12-c",
            "europe-west2-a",
            "europe-west2-b",
            "europe-west2-c",
            "europe-west3-a",
            "europe-west3-b",
            "europe-west3-c",
            "europe-west4-a",
            "europe-west4-b",
            "europe-west4-c",
            "europe-west6-a",
            "europe-west6-b",
            "europe-west6-c",
            "europe-west8-a",
            "europe-west8-b",
            "europe-west8-c",
            "europe-west9-a",
            "europe-west9-b",
            "europe-west9-c",
            "me-central1-a",
            "me-central1-b",
            "me-central1-c",
            "me-central2-a",
            "me-central2-b",
            "me-central2-c",
            "me-west1-a",
            "me-west1-b",
            "me-west1-c",
            "northamerica-northeast1-a",
            "northamerica-northeast1-b",
            "northamerica-northeast1-c",
            "northamerica-northeast2-a",
            "northamerica-northeast2-b",
            "northamerica-northeast2-c",
            "southamerica-east1-a",
            "southamerica-east1-b",
            "southamerica-east1-c",
            "southamerica-west1-a",
            "southamerica-west1-b",
            "southamerica-west1-c",
            "us-central1-a",
            "us-central1-b",
            "us-central1-c",
            "us-central1-f",
            "us-east1-b",
            "us-east1-c",
            "us-east1-d",
            "us-east4-a",
            "us-east4-b",
            "us-east4-c",
            "us-east5-a",
            "us-east5-b",
            "us-east5-c",
            "us-south1-a",
            "us-south1-b",
            "us-south1-c",
            "us-west1-a",
            "us-west1-b",
            "us-west1-c",
            "us-west2-a",
            "us-west2-b",
            "us-west2-c",
            "us-west3-a",
            "us-west3-b",
            "us-west3-c",
            "us-west4-a",
            "us-west4-b",
            "us-west4-c",
        ]
        assert all(f"\ne2-highcpu-2,{i}," in data for i in zones)

    def test_spots_presented(self, data: str):
        assert ",True\n" in data

    def test_ondemand_presented(self, data: str):
        assert ",False\n" in data

    def test_gpu_presented(self, data: str):
        gpus = [
            "A100",
            "L4",
            "T4",
            "V100",
            "P100",
        ]
        assert all(f",{i}," in data for i in gpus)

    def test_both_a100_presented(self, data: str):
        assert ",A100,40.0," in data
        assert ",A100,80.0," in data



import copy
from typing import Optional, Tuple, TypeVar, Union

from gpuhunt._internal.models import CatalogItem, GPUInfo, QueryFilter


def fill_missing(q: QueryFilter, *, memory_per_core: int = 8) -> QueryFilter:
    q = copy.deepcopy(q)

    # if there is some information about gpu
    min_total_gpu_memory = None
    if any(
        value is not None
        for value in (
            q.gpu_name,
            q.min_gpu_count,
            q.min_gpu_memory,
            q.min_total_gpu_memory,
            q.min_compute_capability,
        )
    ):
        if q.min_total_gpu_memory is not None:
            min_total_gpu_memory = q.min_total_gpu_memory
        else:
            min_gpu_count = 1 if q.min_gpu_count is None else q.min_gpu_count
            min_gpu_memory = []
            if q.min_gpu_memory is not None:
                min_gpu_memory.append(q.min_gpu_memory)
            if q.min_compute_capability is not None:
                min_gpu_memory.extend(
                    [
                        i.memory
                        for i in KNOWN_GPUS
                        if i.compute_capability >= q.min_compute_capability
                    ]
                )
            if q.gpu_name is not None:
                min_gpu_memory.extend([i.memory for i in KNOWN_GPUS if i.name in q.gpu_name])
            min_total_gpu_memory = (
                min(min_gpu_memory, default=min(i.memory for i in KNOWN_GPUS)) * min_gpu_count
            )

    if min_total_gpu_memory is not None:
        if q.min_memory is None:  # gpu memory to memory
            q.min_memory = 2 * min_total_gpu_memory
        if q.min_disk_size is None:  # gpu memory to disk
            q.min_disk_size = 30 + min_total_gpu_memory

    if q.min_memory is not None:
        if q.min_cpu is None:  # memory to cpu
            q.min_cpu = (q.min_memory + memory_per_core - 1) // memory_per_core

    if q.min_cpu is not None:
        if q.min_memory is None:  # cpu to memory
            q.min_memory = memory_per_core * q.min_cpu

    return q


Number = TypeVar("Number", bound=Union[int, float])


def optimize(
    available: Number, min_limit: Number, max_limit: Optional[Number]
) -> Optional[Number]:
    if is_above(available, max_limit):
        available = max_limit
    if is_below(available, min_limit):
        return None
    return min_limit


Comparable = TypeVar("Comparable", bound=Union[int, float, Tuple[int, int]])


def is_between(value: Comparable, left: Optional[Comparable], right: Optional[Comparable]) -> bool:
    if is_below(value, left) or is_above(value, right):
        return False
    return True


def is_below(value: Comparable, limit: Optional[Comparable]) -> bool:
    if limit is not None and value < limit:
        return True
    return False


def is_above(value: Comparable, limit: Optional[Comparable]) -> bool:
    if limit is not None and value > limit:
        return True
    return False


def matches(i: CatalogItem, q: QueryFilter) -> bool:
    """
    Check if the catalog item matches the filters

    Args:
        i: catalog item
        q: filters

    Returns:
        whether the catalog item matches the filters
    """
    if q.provider is not None and i.provider not in q.provider:
        return False
    if not is_between(i.cpu, q.min_cpu, q.max_cpu):
        return False
    if not is_between(i.memory, q.min_memory, q.max_memory):
        return False
    if not is_between(i.gpu_count, q.min_gpu_count, q.max_gpu_count):
        return False
    if q.gpu_name is not None:
        if i.gpu_name not in q.gpu_name:
            return False
    if q.min_compute_capability is not None or q.max_compute_capability is not None:
        cc = [info.compute_capability for info in KNOWN_GPUS if info.name == i.gpu_name]
        if not cc or not is_between(min(cc), q.min_compute_capability, q.max_compute_capability):
            return False
    if not is_between(i.gpu_memory if i.gpu_count > 0 else 0, q.min_gpu_memory, q.max_gpu_memory):
        return False
    if not is_between(
        (i.gpu_count * i.gpu_memory) if i.gpu_count > 0 else 0,
        q.min_total_gpu_memory,
        q.max_total_gpu_memory,
    ):
        return False
    # TODO(egor-s): add disk_size to CatalogItem
    # if not is_between(i.disk_size, q.min_disk_size, q.max_disk_size):
    #     return False
    if not is_between(i.price, q.min_price, q.max_price):
        return False
    if q.spot is not None and i.spot != q.spot:
        return False
    return True


KNOWN_GPUS = [
    GPUInfo(name="A10", memory=24, compute_capability=(8, 6)),
    GPUInfo(name="A100", memory=40, compute_capability=(8, 0)),
    GPUInfo(name="A100", memory=80, compute_capability=(8, 0)),
    GPUInfo(name="A10G", memory=24, compute_capability=(8, 6)),
    GPUInfo(name="A4000", memory=16, compute_capability=(8, 6)),
    GPUInfo(name="A6000", memory=48, compute_capability=(8, 6)),
    GPUInfo(name="H100", memory=80, compute_capability=(9, 0)),
    GPUInfo(name="L4", memory=24, compute_capability=(8, 9)),
    GPUInfo(name="L40", memory=48, compute_capability=(8, 9)),
    GPUInfo(name="P100", memory=16, compute_capability=(6, 0)),
    GPUInfo(name="RTX3060Ti", memory=8, compute_capability=(8, 6)),
    GPUInfo(name="RTX3070Ti", memory=8, compute_capability=(8, 6)),
    GPUInfo(name="RTX3080", memory=10, compute_capability=(8, 6)),
    GPUInfo(name="RTX3080Ti", memory=12, compute_capability=(8, 6)),
    GPUInfo(name="RTX3090", memory=24, compute_capability=(8, 6)),
    GPUInfo(name="RTX4090", memory=24, compute_capability=(8, 9)),
    GPUInfo(name="RTX6000", memory=24, compute_capability=(7, 5)),
    GPUInfo(name="T4", memory=16, compute_capability=(7, 5)),
    GPUInfo(name="V100", memory=16, compute_capability=(7, 0)),
    GPUInfo(name="V100", memory=32, compute_capability=(7, 0)),
]



import csv
import dataclasses
import io
import logging
import time
import urllib.request
import zipfile
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import Iterable, List, Optional, Tuple, Union

import gpuhunt._internal.constraints as constraints
from gpuhunt._internal.models import CatalogItem, QueryFilter
from gpuhunt._internal.utils import parse_compute_capability
from gpuhunt.providers import AbstractProvider

logger = logging.getLogger(__name__)
version_url = "https://dstack-gpu-pricing.s3.eu-west-1.amazonaws.com/v1/version"
catalog_url = "https://dstack-gpu-pricing.s3.eu-west-1.amazonaws.com/v1/{version}/catalog.zip"
OFFLINE_PROVIDERS = ["aws", "azure", "gcp", "lambdalabs"]
ONLINE_PROVIDERS = ["tensordock"]
RELOAD_INTERVAL = 4 * 60 * 60  # 4 hours


class Catalog:
    def __init__(self, fill_missing: bool = True, auto_reload: bool = True):
        """
        Args:
            fill_missing: derive missing constraints from other constraints
            auto_reload: if `True`, the catalog will be automatically loaded from the S3 bucket every 4 hours
        """
        self.catalog = None
        self.loaded_at = None
        self.providers: List[AbstractProvider] = []
        self.fill_missing = fill_missing
        self.auto_reload = auto_reload

    def query(
        self,
        *,
        provider: Optional[Union[str, List[str]]] = None,
        min_cpu: Optional[int] = None,
        max_cpu: Optional[int] = None,
        min_memory: Optional[float] = None,
        max_memory: Optional[float] = None,
        min_gpu_count: Optional[int] = None,
        max_gpu_count: Optional[int] = None,
        gpu_name: Optional[Union[str, List[str]]] = None,
        min_gpu_memory: Optional[float] = None,
        max_gpu_memory: Optional[float] = None,
        min_total_gpu_memory: Optional[float] = None,
        max_total_gpu_memory: Optional[float] = None,
        min_disk_size: Optional[int] = None,
        max_disk_size: Optional[int] = None,
        min_price: Optional[float] = None,
        max_price: Optional[float] = None,
        min_compute_capability: Optional[Union[str, Tuple[int, int]]] = None,
        max_compute_capability: Optional[Union[str, Tuple[int, int]]] = None,
        spot: Optional[bool] = None,
    ) -> List[CatalogItem]:
        """
        Query the catalog for matching offers

        Args:
            provider: name of the provider to filter by. If not specified, all providers will be used
            min_cpu: minimum number of CPUs
            max_cpu: maximum number of CPUs
            min_memory: minimum amount of RAM in GB
            max_memory: maximum amount of RAM in GB
            min_gpu_count: minimum number of GPUs
            max_gpu_count: maximum number of GPUs
            gpu_name: case-sensitive name of the GPU to filter by. If not specified, all GPUs will be used
            min_gpu_memory: minimum amount of GPU VRAM in GB for each GPU
            max_gpu_memory: maximum amount of GPU VRAM in GB for each GPU
            min_total_gpu_memory: minimum amount of GPU VRAM in GB for all GPUs combined
            max_total_gpu_memory: maximum amount of GPU VRAM in GB for all GPUs combined
            min_disk_size: *currently not in use*
            max_disk_size: *currently not in use*
            min_price: minimum price per hour in USD
            max_price: maximum price per hour in USD
            min_compute_capability: minimum compute capability of the GPU
            max_compute_capability: maximum compute capability of the GPU
            spot: if `False`, only ondemand offers will be returned. If `True`, only spot offers will be returned

        Returns:
            list of matching offers
        """
        if self.auto_reload and (
            self.loaded_at is None or time.monotonic() - self.loaded_at > RELOAD_INTERVAL
        ):
            self.load()

        query_filter = QueryFilter(
            provider=[provider] if isinstance(provider, str) else provider,
            min_cpu=min_cpu,
            max_cpu=max_cpu,
            min_memory=min_memory,
            max_memory=max_memory,
            min_gpu_count=min_gpu_count,
            max_gpu_count=max_gpu_count,
            gpu_name=[gpu_name] if isinstance(gpu_name, str) else gpu_name,
            min_gpu_memory=min_gpu_memory,
            max_gpu_memory=max_gpu_memory,
            min_total_gpu_memory=min_total_gpu_memory,
            max_total_gpu_memory=max_total_gpu_memory,
            min_disk_size=min_disk_size,
            max_disk_size=max_disk_size,
            min_price=min_price,
            max_price=max_price,
            min_compute_capability=parse_compute_capability(min_compute_capability),
            max_compute_capability=parse_compute_capability(max_compute_capability),
            spot=spot,
        )
        if self.fill_missing:
            query_filter = constraints.fill_missing(query_filter)
        # validate providers
        if provider is not None:
            provider = [p.lower() for p in provider]
            for p in provider:
                if p not in OFFLINE_PROVIDERS + ONLINE_PROVIDERS:
                    raise ValueError(f"Unknown provider: {p}")

        # fetch providers
        items = []
        with ThreadPoolExecutor(max_workers=8) as executor:
            futures = []
            for provider_name in ONLINE_PROVIDERS:
                if query_filter.provider is None or provider_name in query_filter.provider:
                    futures.append(
                        executor.submit(
                            self._get_online_provider_items, provider_name, query_filter
                        )
                    )
            for provider_name in OFFLINE_PROVIDERS:
                if query_filter.provider is None or provider_name in query_filter.provider:
                    futures.append(
                        executor.submit(
                            self._get_offline_provider_items, provider_name, query_filter
                        )
                    )
            for future in as_completed(futures):
                items += future.result()
        return items

    def load(self, version: str = None):
        """
        Fetch the catalog from the S3 bucket

        Args:
            version: specific version of the catalog to download. If not specified, the latest version will be used
        """
        if version is None:
            version = self.get_latest_version()
        logger.debug("Downloading catalog %s...", version)
        with urllib.request.urlopen(catalog_url.format(version=version)) as f:
            self.loaded_at = time.monotonic()
            self.catalog = io.BytesIO(f.read())

    @staticmethod
    def get_latest_version() -> str:
        """
        Get the latest version of the catalog from the S3 bucket
        """
        with urllib.request.urlopen(version_url) as f:
            return f.read().decode("utf-8").strip()

    def add_provider(self, provider: AbstractProvider):
        """
        Add provider for querying offers

        Args:
            provider: provider to add
        """
        self.providers.append(provider)

    def _get_offline_provider_items(
        self, provider_name: str, query_filter: QueryFilter
    ) -> List[CatalogItem]:
        logger.debug("Loading items for offline provider %s", provider_name)
        items = []
        with zipfile.ZipFile(self.catalog) as zip_file:
            with zip_file.open(f"{provider_name}.csv", "r") as csv_file:
                reader: Iterable[dict[str, str]] = csv.DictReader(
                    io.TextIOWrapper(csv_file, "utf-8")
                )
                for row in reader:
                    item = CatalogItem.from_dict(row, provider=provider_name)
                    if constraints.matches(item, query_filter):
                        items.append(item)
        return items

    def _get_online_provider_items(
        self, provider_name: str, query_filter: QueryFilter
    ) -> List[CatalogItem]:
        logger.debug("Loading items for online provider %s", provider_name)
        items = []
        found = False
        for provider in self.providers:
            if provider.NAME != provider_name:
                continue
            found = True
            for i in provider.get(query_filter=query_filter):
                item = CatalogItem(provider=provider_name, **dataclasses.asdict(i))
                if constraints.matches(item, query_filter):
                    items.append(item)
        if not found:
            raise ValueError(f"Provider is not loaded: {provider_name}")
        return items



import copy
import logging
import re
from collections import defaultdict, namedtuple
from typing import List, Optional

import google.cloud.billing_v1 as billing_v1
import google.cloud.compute_v1 as compute_v1

from gpuhunt._internal.models import QueryFilter, RawCatalogItem
from gpuhunt.providers import AbstractProvider

logger = logging.getLogger(__name__)
compute_service = "services/6F81-5844-456A"
AcceleratorDetails = namedtuple("AcceleratorDetails", ["name", "memory"])
accelerator_details = {
    "nvidia-a100-80gb": AcceleratorDetails("A100", 80.0),
    "nvidia-l4": AcceleratorDetails("L4", 24.0),
    "nvidia-tesla-a100": AcceleratorDetails("A100", 40.0),
    "nvidia-tesla-k80": AcceleratorDetails("K80", 12.0),
    "nvidia-tesla-p100": AcceleratorDetails("P100", 16.0),
    "nvidia-tesla-p4": AcceleratorDetails("P4", 8.0),
    "nvidia-tesla-t4": AcceleratorDetails("T4", 16.0),
    "nvidia-tesla-v100": AcceleratorDetails("V100", 16.0),
}
CpuMemory = namedtuple("CpuMemory", ["cpu", "memory"])
accelerator_limits = {
    "nvidia-tesla-k80": [
        CpuMemory(8, 52),
        CpuMemory(16, 104),
        CpuMemory(32, 208),
        CpuMemory(64, 208),
    ],
    "nvidia-tesla-p100": [CpuMemory(16, 104), CpuMemory(32, 208), CpuMemory(96, 624)],
    "nvidia-tesla-p4": [CpuMemory(24, 156), CpuMemory(48, 312), CpuMemory(96, 624)],
    "nvidia-tesla-t4": [CpuMemory(48, 312), CpuMemory(48, 312), CpuMemory(96, 624)],
    "nvidia-tesla-v100": [
        CpuMemory(12, 78),
        CpuMemory(24, 156),
        CpuMemory(48, 312),
        CpuMemory(96, 624),
    ],
}
accelerator_counts = [1, 2, 4, 8, 16]


class GCPProvider(AbstractProvider):
    NAME = "gcp"

    def __init__(self, project: str):
        # todo credentials
        self.project = project
        self.machine_types_client = compute_v1.MachineTypesClient()
        self.accelerator_types_client = compute_v1.AcceleratorTypesClient()
        self.regions_client = compute_v1.RegionsClient()
        self.cloud_catalog_client = billing_v1.CloudCatalogClient()

    def list_preconfigured_instances(self) -> List[RawCatalogItem]:
        instances = []
        for region in self.regions_client.list(project=self.project):
            for zone_url in region.zones:
                zone = zone_url.split("/")[-1]
                logger.info("Fetching instances for zone %s", zone)
                for machine_type in self.machine_types_client.list(
                    project=self.project, zone=zone
                ):
                    if (
                        machine_type.deprecated.state
                        == compute_v1.DeprecationStatus.State.DEPRECATED
                    ):
                        continue
                    gpu = None
                    if machine_type.accelerators:
                        gpu = accelerator_details[
                            machine_type.accelerators[0].guest_accelerator_type
                        ]
                    instance = RawCatalogItem(
                        instance_name=machine_type.name,
                        location=zone,
                        cpu=machine_type.guest_cpus,
                        memory=round(machine_type.memory_mb / 1024, 1),
                        gpu_count=machine_type.accelerators[0].guest_accelerator_count
                        if gpu
                        else 0,
                        gpu_name=machine_type.accelerators[0].guest_accelerator_type
                        if gpu
                        else None,
                        gpu_memory=gpu.memory if gpu else None,
                        price=None,
                        spot=None,
                    )
                    instances.append(instance)
        return instances

    def add_gpus(self, instances: List[RawCatalogItem]):
        n1_instances = defaultdict(list)
        for instance in instances:
            if instance.instance_name.startswith("n1-"):
                n1_instances[instance.location].append(instance)

        instances_with_gpus = []
        for zone, zone_n1_instances in n1_instances.items():
            logger.info("Fetching GPUs for zone %s", zone)
            for accelerator in self.accelerator_types_client.list(project=self.project, zone=zone):
                if accelerator.name not in accelerator_limits:
                    continue
                for n, limit in zip(accelerator_counts, accelerator_limits[accelerator.name]):
                    for instance in zone_n1_instances:
                        if instance.cpu > limit.cpu or instance.memory > limit.memory:
                            continue
                        i = copy.deepcopy(instance)
                        i.gpu_count = n
                        i.gpu_name = accelerator.name
                        i.gpu_memory = accelerator_details[accelerator.name].memory
                        instances_with_gpus.append(i)
        instances += instances_with_gpus

    def fill_prices(self, instances: List[RawCatalogItem]) -> List[RawCatalogItem]:
        logger.info("Fetching prices")
        # fetch per-unit prices
        families = {
            "gpu": defaultdict(dict),
            "ram": defaultdict(dict),
            "core": defaultdict(dict),
        }
        skus = self.cloud_catalog_client.list_skus(parent=compute_service)
        for sku in skus:
            if sku.category.resource_family != "Compute":
                continue
            if sku.category.usage_type not in ["OnDemand", "Preemptible"]:
                continue
            if any(
                word in sku.description
                for word in [
                    "Sole Tenancy",
                    "Reserved",
                    "Premium",
                    "Custom",
                    "suspended",
                ]
            ):
                continue
            r = re.match(
                r"^(?:spot preemptible )?(.+) (gpu|ram|core)",
                sku.description,
                flags=re.IGNORECASE,
            )
            if not r:
                continue

            family, resource = r.groups()
            resource = resource.lower()
            if resource == "gpu":
                family = family.replace(" ", "-").lower()
                family = {"nvidia-tesla-a100-80gb": "nvidia-a100-80gb"}.get(family, family)
            else:
                r = re.match(r"^([a-z]\d.?) ", family.lower())
                if r:
                    family = r.group(1)
                else:
                    family = {
                        "Memory-optimized Instance": "m1",
                        "Compute optimized Instance": "c2",
                        "Compute optimized": "c2",
                    }.get(family, family)

            price = sku.pricing_info[0].pricing_expression.tiered_rates[0].unit_price
            price = price.units + price.nanos / 1e9
            spot = sku.category.usage_type == "Preemptible"
            for region in sku.service_regions:
                families[resource][family][(region, spot)] = price

        # apply per-unit prices to instances
        offers = []
        for instance in instances:
            vm_family = instance.instance_name.split("-")[0]
            if vm_family in [
                "g1",
                "f1",
                "m2",
            ]:  # ignore shared-core and reservation-only
                continue
            for spot in (False, True):
                region_spot = (instance.location[:-2], spot)

                price = 0
                if region_spot not in families["core"][vm_family]:
                    continue
                price += instance.cpu * families["core"][vm_family][region_spot]
                price += instance.memory * families["ram"][vm_family][region_spot]
                if instance.gpu_name:
                    if region_spot not in families["gpu"][instance.gpu_name]:
                        continue
                    price += instance.gpu_count * families["gpu"][instance.gpu_name][region_spot]

                offer = copy.deepcopy(instance)
                offer.price = round(price, 6)
                offer.spot = spot
                if offer.gpu_name:
                    offer.gpu_name = accelerator_details[offer.gpu_name].name
                offers.append(offer)
        return offers

    def get(self, query_filter: Optional[QueryFilter] = None) -> List[RawCatalogItem]:
        instances = self.list_preconfigured_instances()
        self.add_gpus(instances)
        return self.fill_prices(instances)

    @classmethod
    def filter(cls, offers: List[RawCatalogItem]) -> List[RawCatalogItem]:
        return [
            i
            for i in offers
            if any(
                i.instance_name.startswith(family)
                for family in [
                    "e2-medium",
                    "e2-standard-",
                    "e2-highmem-",
                    "e2-highcpu-",
                    "m1-",
                    "a2-",
                    "g2-",
                ]
            )
            or (i.gpu_name and i.gpu_name not in ["K80", "P4"])
        ]



import logging
from typing import List, Optional, Union

import requests

from gpuhunt._internal.constraints import is_between, optimize
from gpuhunt._internal.models import QueryFilter, RawCatalogItem
from gpuhunt.providers import AbstractProvider

logger = logging.getLogger(__name__)

# https://documenter.getpostman.com/view/20973002/2s8YzMYRDc#2b4a3db0-c162-438c-aae4-6a88afc96fdb
marketplace_hostnodes_url = "https://marketplace.tensordock.com/api/v0/client/deploy/hostnodes"
marketplace_gpus = {
    "a100-pcie-80gb": "A100",
    "geforcegtx1070-pcie-8gb": "GTX1070",
    "geforcertx3060-pcie-12gb": "RTX3060",
    "geforcertx3060ti-pcie-8gb": "RTX3060Ti",
    "geforcertx3060tilhr-pcie-8gb": "RTX3060TiLHR",
    "geforcertx3070-pcie-8gb": "RTX3070",
    "geforcertx3070ti-pcie-8gb": "RTX3070Ti",
    "geforcertx3080-pcie-10gb": "RTX3080",
    "geforcertx3080ti-pcie-12gb": "RTX3080Ti",
    "geforcertx3090-pcie-24gb": "RTX3090",
    "geforcertx4090-pcie-24gb": "RTX4090",
    "l40-pcie-48gb": "L40",
    "rtxa4000-pcie-16gb": "A4000",
    "rtxa5000-pcie-24gb": "A5000",
    "rtxa6000-pcie-48gb": "A6000",
    "v100-pcie-16gb": "V100",
}


class TensorDockProvider(AbstractProvider):
    NAME = "tensordock"

    def get(self, query_filter: Optional[QueryFilter] = None) -> List[RawCatalogItem]:
        logger.info("Fetching TensorDock offers")
        hostnodes = requests.get(marketplace_hostnodes_url).json()["hostnodes"]
        offers = []
        for hostnode, details in hostnodes.items():
            location = (
                "-".join([details["location"][key] for key in ["country", "region", "city"]])
                .lower()
                .replace(" ", "")
            )
            if query_filter is not None:
                offers += self.optimize_offers(query_filter, details["specs"], hostnode, location)
            else:
                for gpu_name, gpu in details["specs"]["gpu"].items():
                    if gpu["amount"] == 0:
                        continue
                    offers.append(
                        RawCatalogItem(
                            instance_name=hostnode,
                            location=location,
                            price=round(
                                sum(
                                    details["specs"][key]["price"]
                                    * details["specs"][key]["amount"]
                                    for key in ("cpu", "ram", "storage")
                                )
                                + gpu["amount"] * gpu["price"],
                                5,
                            ),
                            cpu=details["specs"]["cpu"]["amount"],
                            memory=float(round_down(details["specs"]["ram"]["amount"], 2)),
                            gpu_count=gpu["amount"],
                            gpu_name=marketplace_gpus.get(gpu_name, gpu_name),
                            gpu_memory=float(gpu["vram"]),
                            spot=False,
                        )
                    )
        return offers

    @staticmethod
    def optimize_offers(
        q: QueryFilter, specs: dict, instance_name: str, location: str
    ) -> List[RawCatalogItem]:
        cpu = optimize(specs["cpu"]["amount"], q.min_cpu or 1, q.max_cpu)
        memory = optimize(  # has to be even
            round_down(specs["ram"]["amount"], 2),
            round_up(q.min_memory or 1, 2),
            round_down(q.max_memory, 2) if q.max_memory is not None else None,
        )
        disk_size = optimize(  # 30 GB at least for Ubuntu
            specs["storage"]["amount"],
            q.min_disk_size or 30,
            q.max_disk_size,
        )
        if cpu is None or memory is None or disk_size is None:
            return []
        base_price = sum(
            n * specs[key]["price"]
            for key, n in [("cpu", cpu), ("ram", memory), ("storage", disk_size)]
        )
        offers = []
        for gpu_name, gpu in specs["gpu"].items():
            gpu_name = marketplace_gpus.get(gpu_name, gpu_name)
            if q.gpu_name is not None and gpu_name not in q.gpu_name:
                continue
            if not is_between(gpu["vram"], q.min_gpu_memory, q.max_gpu_memory):
                continue
            if (
                gpu_count := optimize(gpu["amount"], q.min_gpu_count or 1, q.max_gpu_count)
            ) is None:
                continue
            # filter by total gpu memory
            if q.min_total_gpu_memory is None:
                min_total_gpu_memory = gpu_count * gpu["vram"]
            else:
                min_total_gpu_memory = max(q.min_total_gpu_memory, gpu_count * gpu["vram"])
            gpu_total_memory = optimize(
                gpu["amount"] * gpu["vram"],
                round_up(min_total_gpu_memory, gpu["vram"]),
                round_down(q.max_total_gpu_memory, gpu["vram"])
                if q.max_total_gpu_memory is not None
                else None,
            )
            if gpu_total_memory is None:
                continue
            gpu_count = gpu_total_memory // gpu["vram"]
            if not is_between(gpu_count, q.min_gpu_count, q.max_gpu_count):
                continue
            # make an offer
            offer = RawCatalogItem(
                instance_name=instance_name,
                location=location,
                price=round(gpu_count * gpu["price"] + base_price, 5),
                cpu=cpu,
                memory=float(memory),
                gpu_count=gpu_count,
                gpu_name=gpu_name,
                gpu_memory=float(gpu["vram"]),
                spot=False,
            )
            offers.append(offer)
        return offers


def round_up(value: Union[int, float], step: int) -> int:
    return round_down(value + step - 1, step)


def round_down(value: Union[int, float], step: int) -> int:
    return value // step * step
